package com.example.paybyworkout.classifier;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import android.content.ContentValues;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;
import android.util.Log;
import android.widget.TextView;

import com.example.paybyworkout.WorkoutActivity;
import com.example.paybyworkout.R;
import com.example.paybyworkout.classes.ClassificationFeatures;
import com.example.paybyworkout.classes.WorkoutMode;
import com.example.paybyworkout.database.MySQLiteHelper;
import com.example.paybyworkout.featureextraction.FeatureExtractionThread;

/**
 * Uses Naive Bayes algorithm for classification. Reads the classification features instance and classifies it by the given model. 
 * Afterwards the classification result gets stored in the database and the UI of the workout fragment gets updated.
 * @author Jan
 */
public class NBClassifierThread extends Thread {

	private WorkoutActivity activity;
	private ClassificationFeatures classificationFeatures;
	private double classificationSum = 0;
	private Classifier model;
	private Instances trainingSet;

	/**
	 * Creates new NBClassifierThread, which uses the Naive Bayes algorithm
	 * @param activity WorkoutActivity, which gets updated with the classification results
	 * @param classificationFeatures The latest computed features, which are used for classification
	 * @param model Contains a Naive Bayes model, trained with a training set
	 * @param trainingSet The training set used for evaluation
	 */
	public NBClassifierThread(WorkoutActivity activity, ClassificationFeatures classificationFeatures, Classifier model, Instances trainingSet) {
		this.activity = activity;
		this.classificationFeatures = classificationFeatures;
		this.model = model;
		this.trainingSet = trainingSet;	
		
	}

	public void run() {	

		try {

			Evaluation testRunning = new Evaluation(trainingSet);		

			Instances testingInstance = new Instances("Evaluation", getAttributes(), 1);
			testingInstance.setClassIndex(0);
		
			classificationFeatures.setClassification(WorkoutMode.RUNNING);
			testingInstance.add(getInstanceFromClassificationFeatures(classificationFeatures));
			testRunning.evaluateModel(model, testingInstance);

			double prediction = testRunning.correct();

			//Insert into database
			ContentValues contentValues = new ContentValues();
			contentValues.put(MySQLiteHelper.CLASSIFICATION_COLUMN_TIMESTAMP, System.currentTimeMillis());
			contentValues.put(MySQLiteHelper.CLASSIFICATION_COLUMN_RESULT, String.valueOf(prediction));	
			activity.getDatabase().insert(MySQLiteHelper.TABLE_CLASSIFICATION_DATA, null, contentValues);
			
			Log.d("NBClassifierThread", "Classification: " + prediction + " with values " + classificationFeatures.toString());
			
			//Get average classification within last window
			long minTime = System.currentTimeMillis() - FeatureExtractionThread.CLASSIFICATION_WINDOW_SIZE;

			SQLiteDatabase db = activity.getSqlHelper().getReadableDatabase();
			Cursor res =  db.rawQuery( "SELECT * FROM " + MySQLiteHelper.TABLE_CLASSIFICATION_DATA
					+ " WHERE "
					+ MySQLiteHelper.CLASSIFICATION_COLUMN_TIMESTAMP + " > " + minTime + " AND "
					+ MySQLiteHelper.CLASSIFICATION_COLUMN_TIMESTAMP + " < " + System.currentTimeMillis(), null );
			res.moveToFirst();

			while(res.isAfterLast() == false){
				classificationSum += Double.valueOf(res.getString(2));
				res.moveToNext();
			}
			
			Log.d("NBClassifierThread", "classificationSum: " + classificationSum + " Elapsed time: " + (System.currentTimeMillis() - activity.getStartRecordingTime()));
			
			activity.runOnUiThread(new Runnable() {
				public void run() {
					TextView statusText = ((TextView)activity.findViewById(R.id.status));
					TextView timeText = ((TextView)activity.findViewById(R.id.time));

					if(classificationSum >= 3 || (System.currentTimeMillis() - activity.getStartRecordingTime() < FeatureExtractionThread.CLASSIFICATION_WINDOW_SIZE)) {
						timeText.setTextColor(activity.getResources().getColor(R.color.workoutRunning));
						statusText.setTextColor(activity.getResources().getColor(R.color.workoutRunning));
						if(activity.getWorkoutMode().equals(WorkoutMode.RUNNING)) {
							statusText.setText(activity.getResources().getString(R.string.running));
						}else if(activity.getWorkoutMode().equals(WorkoutMode.CYCLING)) {
							statusText.setText(activity.getResources().getString(R.string.cycling));					
						}

					}else{
						timeText.setTextColor(activity.getResources().getColor(R.color.workoutOther));
						statusText.setTextColor(activity.getResources().getColor(R.color.workoutOther));
						if(activity.getWorkoutMode().equals(WorkoutMode.RUNNING)) {
							statusText.setText(activity.getResources().getString(R.string.not_running));
						}else if(activity.getWorkoutMode().equals(WorkoutMode.CYCLING)) {
							statusText.setText(activity.getResources().getString(R.string.not_cycling));					
						}
					}
				}
			});
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public static Instance getInstanceFromClassificationFeatures(ClassificationFeatures classificationFeatures) {
		Instance classificationInstance = new Instance(32);
		classificationInstance.setValue(1, classificationFeatures.getAccelerationAverageX());
		classificationInstance.setValue(2, classificationFeatures.getAccelerationAverageY());
		classificationInstance.setValue(3, classificationFeatures.getAccelerationAverageZ());
		classificationInstance.setValue(4, classificationFeatures.getAccelerationXYcorrelation());
		classificationInstance.setValue(5, classificationFeatures.getAccelerationXZcorrelation());
		classificationInstance.setValue(6, classificationFeatures.getAccelerationYZcorrelation());
		classificationInstance.setValue(7, classificationFeatures.getxVariance());
		classificationInstance.setValue(8, classificationFeatures.getyVariance());
		classificationInstance.setValue(9, classificationFeatures.getzVariance());
		classificationInstance.setValue(10, classificationFeatures.getxMin());
		classificationInstance.setValue(11, classificationFeatures.getxMax());
		classificationInstance.setValue(12, classificationFeatures.getyMin());
		classificationInstance.setValue(13, classificationFeatures.getyMax());
		classificationInstance.setValue(14, classificationFeatures.getzMin());
		classificationInstance.setValue(15, classificationFeatures.getzMax());
		classificationInstance.setValue(16, classificationFeatures.getAccelerationAverageXFFT());
		classificationInstance.setValue(17, classificationFeatures.getAccelerationAverageYFFT());
		classificationInstance.setValue(18, classificationFeatures.getAccelerationAverageZFFT());
		classificationInstance.setValue(19, classificationFeatures.getAccelerationXYcorrelationFFT());
		classificationInstance.setValue(20, classificationFeatures.getAccelerationXZcorrelationFFT());
		classificationInstance.setValue(21, classificationFeatures.getAccelerationYZcorrelationFFT());
		classificationInstance.setValue(22, classificationFeatures.getxVarianceFFT());
		classificationInstance.setValue(23, classificationFeatures.getyVarianceFFT());
		classificationInstance.setValue(24, classificationFeatures.getzVarianceFFT());
		classificationInstance.setValue(25, classificationFeatures.getxMinFFT());
		classificationInstance.setValue(26, classificationFeatures.getxMaxFFT());
		classificationInstance.setValue(27, classificationFeatures.getyMinFFT());
		classificationInstance.setValue(28, classificationFeatures.getyMaxFFT());
		classificationInstance.setValue(29, classificationFeatures.getzMinFFT());
		classificationInstance.setValue(30, classificationFeatures.getzMaxFFT());
		classificationInstance.setValue(31, classificationFeatures.getAverageSpeed());

		if(classificationFeatures.getClassification().equals(WorkoutMode.RUNNING)) {
			classificationInstance.setValue(0, 1);
		}else if(classificationFeatures.getClassification().equals(WorkoutMode.CYCLING)) {
			classificationInstance.setValue(0, 1);
		}else{
			classificationInstance.setValue(0, 0);				
		}

		return classificationInstance;
	}

	public static FastVector getAttributes() {
		FastVector attributes = new FastVector(32);
		//Defining class
		FastVector attributeClass = new FastVector(2);
		attributeClass.addElement("0");
		attributeClass.addElement("1");

		attributes.addElement(new Attribute("classification", attributeClass));
		attributes.addElement(new Attribute("accelerationAverageX"));
		attributes.addElement(new Attribute("accelerationAverageY"));
		attributes.addElement(new Attribute("accelerationAverageZ"));
		attributes.addElement(new Attribute("accelerationXYcorrelation"));
		attributes.addElement(new Attribute("accelerationXZcorrelation"));
		attributes.addElement(new Attribute("accelerationYZcorrelation"));
		attributes.addElement(new Attribute("xVariance"));
		attributes.addElement(new Attribute("yVariance"));
		attributes.addElement(new Attribute("zVariance"));
		attributes.addElement(new Attribute("xMin"));
		attributes.addElement(new Attribute("xMax"));
		attributes.addElement(new Attribute("yMin"));
		attributes.addElement(new Attribute("yMax"));
		attributes.addElement(new Attribute("zMin"));
		attributes.addElement(new Attribute("zMax"));
		attributes.addElement(new Attribute("accelerationAverageXFFT"));
		attributes.addElement(new Attribute("accelerationAverageYFFT"));
		attributes.addElement(new Attribute("accelerationAverageZFFT"));
		attributes.addElement(new Attribute("accelerationXYcorrelationFFT"));
		attributes.addElement(new Attribute("accelerationXZcorrelationFFT"));
		attributes.addElement(new Attribute("accelerationYZcorrelationFFT"));
		attributes.addElement(new Attribute("xVarianceFFT"));
		attributes.addElement(new Attribute("yVarianceFFT"));
		attributes.addElement(new Attribute("zVarianceFFT"));
		attributes.addElement(new Attribute("xMinFFT"));
		attributes.addElement(new Attribute("xMaxFFT"));
		attributes.addElement(new Attribute("yMinFFT"));
		attributes.addElement(new Attribute("yMaxFFT"));
		attributes.addElement(new Attribute("zMinFFT"));
		attributes.addElement(new Attribute("zMaxFFT"));
		attributes.addElement(new Attribute("averageSpeed"));

		return attributes;
	}
}
