package com.example.paybyworkout.featureextraction;

import java.io.BufferedReader;
import java.io.InputStreamReader;

import org.apache.commons.math3.complex.Complex;
import org.apache.commons.math3.stat.correlation.PearsonsCorrelation;
import org.apache.commons.math3.transform.DftNormalization;
import org.apache.commons.math3.transform.FastFourierTransformer;
import org.apache.commons.math3.transform.TransformType;

import weka.classifiers.Classifier;
import weka.classifiers.bayes.NaiveBayes;
import weka.core.Instances;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;
import android.util.Log;

import com.example.paybyworkout.R;
import com.example.paybyworkout.WorkoutActivity;
import com.example.paybyworkout.classes.ClassificationFeatures;
import com.example.paybyworkout.classes.WorkoutMode;
import com.example.paybyworkout.classifier.LibSVMClassifierThread;
import com.example.paybyworkout.classifier.NBClassifierThread;
import com.example.paybyworkout.database.MySQLiteHelper;

/**
 * Gets latest features within specified windows out of the recorded data in the database and executes the classification of this data afterwards.
 * @author Jan
 */
public class FeatureExtractionThread extends Thread {

	public static final boolean LIBSVM_CLASSIFIER = false;
	
	public static final long ACCELERATION_WINDOW_SIZE = 5000;
	public static final long GPS_WINDOW_SIZE = 20000;
	public static final long CLASSIFICATION_INTERVAL = 2000;
	public static final long CLASSIFICATION_WINDOW_SIZE = 30000;

	private WorkoutActivity activity;

	private ClassificationFeatures classificationFeatures = new ClassificationFeatures();

	public FeatureExtractionThread(WorkoutActivity activity) {
		this.activity = activity;
	}

	public void run() {
		Classifier model = (Classifier) new NaiveBayes();
		Instances trainingSet = null;		
		
		if(!LIBSVM_CLASSIFIER) {
			//Prepare training data for NB
			
			BufferedReader reader = null;
			
			try {
	
				if(activity.getWorkoutMode().equals(WorkoutMode.RUNNING)) {
					reader = new BufferedReader(new InputStreamReader(activity.getResources().openRawResource(R.raw.nb_train_running)));
				}else if(activity.getWorkoutMode().equals(WorkoutMode.CYCLING)) {
					reader = new BufferedReader(new InputStreamReader(activity.getResources().openRawResource(R.raw.nb_train_cycling)));				
				}
				
				trainingSet = new Instances(reader);
				reader.close();
				// setting class attribute
				trainingSet.setClassIndex(0);			
	
				model.buildClassifier(trainingSet);				
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		while(activity.isRecording()) {

			//Read data out of database into variables
			updateAccelerationFeatures();				
			updateGPSFeatures();			

			//Classify data afterwards
			if(LIBSVM_CLASSIFIER) {
				LibSVMClassifierThread classifier = new LibSVMClassifierThread(activity, classificationFeatures);
				classifier.start();
			}else{
				NBClassifierThread classifier = new NBClassifierThread(activity, classificationFeatures, model, trainingSet);
				classifier.start();
			}

			try {
				Thread.sleep(CLASSIFICATION_INTERVAL);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

	private void updateAccelerationFeatures() {
		//Read acceleration data
		long minTime = System.currentTimeMillis() - ACCELERATION_WINDOW_SIZE;

		SQLiteDatabase db = activity.getSqlHelper().getReadableDatabase();
		Cursor res =  db.rawQuery( "SELECT * FROM " + MySQLiteHelper.TABLE_ACCELERATION_DATA
				+ " WHERE "
				+ MySQLiteHelper.ACCELERATION_COLUMN_TIMESTAMP + " > " + minTime + " AND "
				+ MySQLiteHelper.ACCELERATION_COLUMN_TIMESTAMP + " < " + System.currentTimeMillis(), null );
		res.moveToFirst();

		double[] xValue = new double[res.getCount()];
		double[] yValue = new double[res.getCount()];
		double[] zValue = new double[res.getCount()];

		int i = 0;
		while(res.isAfterLast() == false){
			xValue[i] = Double.valueOf(res.getString(2));
			yValue[i] = Double.valueOf(res.getString(3));
			zValue[i] = Double.valueOf(res.getString(4));
			res.moveToNext();
			i++;
		}

		//Read GPS data
		float[] speed = new float[0];		
		classificationFeatures = getFeaturesOutOfData(xValue, yValue, zValue, speed);		
	}

	public static ClassificationFeatures getFeaturesOutOfData(double[] xValue, double[] yValue, double[] zValue, float[] speed) {

		//Variables
		ClassificationFeatures classificationFeatures = new ClassificationFeatures();
		float accelerationSumX = 0;
		float accelerationSumY = 0;
		float accelerationSumZ = 0;
		float accelerationCount = 0;

		float xAverage = 0;
		float yAverage = 0;
		float zAverage = 0;
		float xVariance = 0;
		float yVariance = 0;
		float zVariance = 0;

		double xMin = Double.MAX_VALUE;
		double xMax = 0;
		double yMin = Double.MAX_VALUE;
		double yMax = 0;			
		double zMin = Double.MAX_VALUE;
		double zMax = 0;

		double[] xValueFFT;
		double[] yValueFFT;
		double[] zValueFFT;

		float averageSpeed = 0;

		if(xValue.length == 0) {
			return classificationFeatures;
		}
		
		for(int i = 0; i < xValue.length; i++) {
			accelerationSumX += xValue[i];
			accelerationSumY += yValue[i];
			accelerationSumZ += zValue[i];
			accelerationCount++;

			xAverage += xValue[i];
			yAverage += yValue[i];
			zAverage += zValue[i];

			if(xValue[i] < xMin) xMin = xValue[i];
			if(xValue[i] > xMax) xMax = xValue[i];
			if(yValue[i] < yMin) yMin = yValue[i];
			if(yValue[i] > yMax) yMax = yValue[i];
			if(zValue[i] < zMin) zMin = zValue[i];
			if(zValue[i] > zMax) zMax = zValue[i];
		}

		classificationFeatures.setAccelerationAverageX(accelerationSumX/accelerationCount);
		classificationFeatures.setAccelerationAverageY(accelerationSumY/accelerationCount);
		classificationFeatures.setAccelerationAverageZ(accelerationSumZ/accelerationCount);
		PearsonsCorrelation correlation = new PearsonsCorrelation();
		classificationFeatures.setAccelerationXYcorrelation(correlation.correlation(xValue, yValue));
		classificationFeatures.setAccelerationXZcorrelation(correlation.correlation(xValue, zValue));
		classificationFeatures.setAccelerationYZcorrelation(correlation.correlation(yValue, zValue));					
		classificationFeatures.setxMin(xMin);
		classificationFeatures.setxMax(xMax);
		classificationFeatures.setyMin(yMin);
		classificationFeatures.setyMax(yMax);
		classificationFeatures.setzMin(zMin);
		classificationFeatures.setzMax(zMax);

		//Get variance
		xAverage /= xValue.length;
		yAverage /= xValue.length;
		zAverage /= xValue.length;
		for(int i = 0; i < xValue.length; i++) {
			xVariance += Math.pow(xAverage-xValue[i], 2);
			yVariance += Math.pow(yAverage-yValue[i], 2);
			zVariance += Math.pow(zAverage-zValue[i], 2);
		}	
		xVariance /= xValue.length;
		yVariance /= xValue.length;
		yVariance /= xValue.length;

		classificationFeatures.setxVariance(xVariance);
		classificationFeatures.setyVariance(yVariance);
		classificationFeatures.setzVariance(zVariance);

		//Get FFT
		FastFourierTransformer fft = new FastFourierTransformer(DftNormalization.STANDARD);

		//Modifying array size to the power of 2
		int sizeForFFT = getNextPowerOfTwo(xValue.length);
		xValueFFT = new double[sizeForFFT];
		yValueFFT = new double[sizeForFFT];
		zValueFFT = new double[sizeForFFT];
		System.arraycopy(xValue, 0, xValueFFT, 0, sizeForFFT);
		System.arraycopy(yValue, 0, yValueFFT, 0, sizeForFFT);
		System.arraycopy(zValue, 0, zValueFFT, 0, sizeForFFT);

		Complex[] xFFT = fft.transform(xValueFFT, TransformType.FORWARD);
		Complex[] yFFT = fft.transform(yValueFFT, TransformType.FORWARD);
		Complex[] zFFT = fft.transform(zValueFFT, TransformType.FORWARD);
		xValueFFT = new double[xFFT.length];
		yValueFFT = new double[yFFT.length];
		zValueFFT = new double[zFFT.length];
		for(int i = 0; i < xFFT.length; i++) {
			xValueFFT[i] = xFFT[i].getReal();
			yValueFFT[i] = yFFT[i].getReal();
			zValueFFT[i] = zFFT[i].getReal();
		}

		float accelerationSumXFFT = 0;
		float accelerationSumYFFT = 0;
		float accelerationSumZFFT = 0;
		float accelerationCountFFT = 0;

		float xAverageFFT = 0;
		float yAverageFFT = 0;
		float zAverageFFT = 0;
		float xVarianceFFT = 0;
		float yVarianceFFT = 0;
		float zVarianceFFT = 0;

		double xMinFFT = Double.MAX_VALUE;
		double xMaxFFT = 0;
		double yMinFFT = Double.MAX_VALUE;
		double yMaxFFT = 0;			
		double zMinFFT = Double.MAX_VALUE;
		double zMaxFFT = 0;
		
		for(int i = 0; i < xValueFFT.length; i++) {
			accelerationSumXFFT += xValueFFT[i];
			accelerationSumYFFT += yValueFFT[i];			
			accelerationSumZFFT += zValueFFT[i];
			accelerationCountFFT++;

			xAverageFFT += xValueFFT[i];
			yAverageFFT += yValueFFT[i];
			zAverageFFT += zValueFFT[i];

			if(xValueFFT[i] < xMinFFT) xMinFFT = xValueFFT[i];
			if(xValueFFT[i] > xMaxFFT) xMaxFFT = xValueFFT[i];
			if(yValueFFT[i] < yMinFFT) yMinFFT = yValueFFT[i];
			if(yValueFFT[i] > yMaxFFT) yMaxFFT = yValueFFT[i];
			if(zValueFFT[i] < zMinFFT) zMinFFT = zValueFFT[i];
			if(zValueFFT[i] > zMaxFFT) zMaxFFT = zValueFFT[i];
		}

		classificationFeatures.setAccelerationAverageXFFT(accelerationSumXFFT/accelerationCountFFT);
		classificationFeatures.setAccelerationAverageYFFT(accelerationSumYFFT/accelerationCountFFT);
		classificationFeatures.setAccelerationAverageZFFT(accelerationSumZFFT/accelerationCountFFT);
		PearsonsCorrelation correlationFFT = new PearsonsCorrelation();
		classificationFeatures.setAccelerationXYcorrelationFFT(correlationFFT.correlation(xValueFFT, yValueFFT));
		classificationFeatures.setAccelerationXZcorrelationFFT(correlationFFT.correlation(xValueFFT, zValueFFT));
		classificationFeatures.setAccelerationYZcorrelationFFT(correlationFFT.correlation(yValueFFT, zValueFFT));					
		classificationFeatures.setxMinFFT(xMinFFT);
		classificationFeatures.setxMaxFFT(xMaxFFT);
		classificationFeatures.setyMinFFT(yMinFFT);
		classificationFeatures.setyMaxFFT(yMaxFFT);
		classificationFeatures.setzMinFFT(zMinFFT);
		classificationFeatures.setzMaxFFT(zMaxFFT);

		//Get variance
		xAverageFFT /= xValueFFT.length;
		yAverageFFT /= xValueFFT.length;
		zAverageFFT /= xValueFFT.length;
		for(int i = 0; i < xValueFFT.length; i++) {
			xVarianceFFT += Math.pow(xAverageFFT-xValueFFT[i], 2);
			yVarianceFFT += Math.pow(yAverageFFT-yValueFFT[i], 2);
			zVarianceFFT += Math.pow(zAverageFFT-zValueFFT[i], 2);
		}	
		xVarianceFFT /= xValueFFT.length;
		yVarianceFFT /= xValueFFT.length;
		yVarianceFFT /= xValueFFT.length;

		classificationFeatures.setxVarianceFFT(xVarianceFFT);
		classificationFeatures.setyVarianceFFT(yVarianceFFT);
		classificationFeatures.setzVarianceFFT(zVarianceFFT);		
		
		
		if(speed.length != 0) {
			//Set speed
			for(int i = 0; i < speed.length; i++) {
				averageSpeed += speed[i];
			}
			averageSpeed /= speed.length;
			classificationFeatures.setAverageSpeed(averageSpeed);
		}

		return classificationFeatures;
	}

	private static int getNextPowerOfTwo(int n)	{
		//If n is already power of 2
		if ((n & (n-1)) == 0) {
			return (n);
		}

		while (!((n & (n-1)) == 0)) {
			n = n & (n-1);
		}
		return n;
	}

	private void updateGPSFeatures() {
		long minTime = System.currentTimeMillis() - GPS_WINDOW_SIZE;

		float sum = 0;
		float count = 0;

		SQLiteDatabase db = activity.getSqlHelper().getReadableDatabase();
		Cursor res =  db.rawQuery( "SELECT * FROM " + MySQLiteHelper.TABLE_GPS_DATA
				+ " WHERE "
				+ MySQLiteHelper.GPS_COLUMN_TIMESTAMP + " > " + minTime + " AND "
				+ MySQLiteHelper.GPS_COLUMN_TIMESTAMP + " < " + System.currentTimeMillis(), null );
		res.moveToFirst();

		while(res.isAfterLast() == false){
			sum += res.getFloat(5);
			count++;
			res.moveToNext();
		}

		if(count > 0) {
			classificationFeatures.setAverageSpeed(sum/count);
		}else{
			Log.d("FeatureExtractionThread", "Currently no GPS data in database");		
		}
	}


}
